function cost_multiscale = WaveletBasedGeodesicsCompare(im1, im2)
% Returns the similarity cost between two images as defined in:
% A. Jorstad, D. Jacobs, A. Trouve. "A Fast Illumination and Deformation
% Insensitive Image Comparison Algorithm Using Wavelet-Based Geodesics." 
% European Conference on Computer Vision (ECCV), Oct. 2012.
%
% The file WaveletBasedGeodesicsLookupTable.mat (generated by 
% generate_lookup_table.m) must be present in the current directory.
%
% When comparing many images, it is suggested that you precompute all 
% wavelet coefficients first (the first for-loop), then run the following 
% lookup loop.
%
% Contact:  anne.jorstad@epfl.ch
%
% Copyright (c) 2012, Anne Jorstad
% May be used for academic purposes


if size(im1) ~= size(im2)
    disp('images must be of the same dimension')
    cost_multiscale = -1;
    return
end
if size(im1,3) ~= 1
    disp('images must be 2D')
    cost_multiscale = -1;
    return
end

nscls = 3;
wav_type = 'bior1.3';
load('WaveletBasedGeodesicsLookupTable.mat')  
% loads r_rng, tdiff_rng, costs_table


% manipulations to aid in vector calculations later
costs_table_vec = costs_table(:);
nr = length(r_rng);
r_rng3_sub = ones(1,1,nr);
r_rng3_sub(1,1,:) = r_rng;
ntd = length(tdiff_rng);
tdiff_rng3_sub = ones(1,1,ntd);
tdiff_rng3_sub(1,1,:) = tdiff_rng;
tdiff_rng3 = cell(1,nscls);


% precalculate wavelet coefficients and convert to polar (rs and ts);
% this significantly increases efficiency when comparing many images
% (pre-compute, then feed ts and inds_rs to comparison step)
ts = cell(2, nscls);  % theta values
inds_rs = cell(2, nscls);  % indices of radius values in the lookup table
for scl = 1:nscls
    % get coefficient locations
    [all_coefs1, all_sizes] = wavedec2(im1, scl, wav_type);
    cH_start = prod(all_sizes(1,1:2)) + 1;
    cH_end = cH_start + prod(all_sizes(2,1:2)) - 1;
    cV_start = prod(all_sizes(1,1:2)) + prod(all_sizes(2,1:2)) + 1;
    cV_end = cV_start + prod(all_sizes(2,1:2)) - 1;
    w_len = all_sizes(2,1);
    w_wid = all_sizes(2,2);
    r_rng3 = repmat(r_rng3_sub, w_len, w_wid);
    tdiff_rng3{scl} = repmat(tdiff_rng3_sub, w_len, w_wid);

    % extract coefficients for im1
    cH1_vec = all_coefs1(cH_start:cH_end);
    cH1 = reshape(cH1_vec, w_len, w_wid);
    cV1_vec = all_coefs1(cV_start:cV_end);
    cV1 = reshape(cV1_vec, w_len, w_wid);
    r1 = sqrt(cH1.^2 + cV1.^2);
    t1 = atan2(cV1, cH1);
    r1_here = repmat(r1, [1 1 nr]);
    diffs_r1 = abs(r1_here - r_rng3);
    [~, inds_r1] = min(diffs_r1, [], 3);

    % extract coefficients for im2
    [all_coefs2, ~] = wavedec2(im2, scl, wav_type);
    cH2_vec = all_coefs2(cH_start:cH_end);
    cH2 = reshape(cH2_vec, w_len, w_wid);
    cV2_vec = all_coefs2(cV_start:cV_end);
    cV2 = reshape(cV2_vec, w_len, w_wid);
    r2 = sqrt(cH2.^2 + cV2.^2);
    t2 = atan2(cV2, cH2);
    r2_here = repmat(r2, [1 1 nr]);
    diffs_r2 = abs(r2_here - r_rng3);
    [~, inds_r2] = min(diffs_r2, [], 3);

    % store
    ts{1,scl} = t1;
    ts{2,scl} = t2;
    inds_rs{1,scl} = inds_r1;
    inds_rs{2,scl} = inds_r2;
end


% lookup the point-based costs at each scale
costs = zeros(1,nscls);
for scl = 1:nscls
    t1 = ts{1,scl};
    t2 = ts{2,scl};
    inds_r1 = inds_rs{1,scl};
    inds_r2 = inds_rs{2,scl};

    % calculate the difference in thetas at every point
    tdiff = mod(abs(t2 - t1), 2*pi);  % 2*pi, not just pi
    tdiff(tdiff>pi) = 2*pi - tdiff(tdiff>pi);
    tdiff_here = repmat(tdiff, [1 1 ntd]);
    diffs_tdiff = abs(tdiff_here - tdiff_rng3{scl});
    [~, inds_tdiff] = min(diffs_tdiff, [], 3);  % indices of tdiff values 
                                                % in the lookup table
        
    inds_mat = sub2ind([nr nr ntd], inds_r1, inds_r2, inds_tdiff);
    costs_here = costs_table_vec(inds_mat);
    costs(scl) = single(sum(costs_here(:)));  % = norm(costs_here(:), 1);
end


% combine costs at all scales
cost_multiscale = 0;
for scl = 1:nscls
    disp(costs(scl))
    coef = 2^scl;
    cost_multiscale = cost_multiscale + coef * costs(scl);
end



